#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import datetime
import hashlib
import inspect
import json
import os

from enum import IntEnum
from functools import wraps
from os import getcwd, getenv, makedirs
from pathlib import Path
from platform import python_version
from sys import platform
from typing import Any, Callable, Optional, TypeVar
from uuid import getnode

from snowflake.connector import (
    SNOWFLAKE_CONNECTOR_VERSION,
    time_util,
)
from snowflake.connector.constants import DIRS as SNOWFLAKE_DIRS
from snowflake.connector.network import SnowflakeRestful
from snowflake.connector.telemetry import TelemetryClient
from snowflake.snowpark import VERSION as SNOWPARK_VERSION
from snowflake.snowpark import dataframe as snowpark_dataframe
from snowflake.snowpark.session import Session


try:
    from pyspark.sql import dataframe as spark_dataframe

    def _is_spark_dataframe(df: Any) -> bool:
        return isinstance(df, spark_dataframe.DataFrame)

    def _get_spark_schema_types(df: spark_dataframe.DataFrame) -> list[str]:
        return [str(schema_type.dataType) for schema_type in df.schema.fields]

except Exception:

    def _is_spark_dataframe(df: Any):
        pass

    def _get_spark_schema_types(df: Any):
        pass


class TelemetryManager(TelemetryClient):
    def __init__(self, rest: SnowflakeRestful):
        """TelemetryManager class to log telemetry events."""
        super().__init__(rest)
        self.sc_folder_path = (
            Path(SNOWFLAKE_DIRS.user_config_path) / "snowpark-checkpoints-telemetry"
        )
        self.sc_sf_path_telemetry = "/telemetry/send"
        self.sc_flush_size = 25
        self.sc_is_enabled = self._sc_is_telemetry_enabled()
        self.sc_is_testing = self._sc_is_telemetry_testing()
        self.sc_memory_limit = 5 * 1024 * 1024
        self._sc_upload_local_telemetry()
        self.sc_log_batch = []
        self.sc_hypothesis_input_events = []

    def set_sc_output_path(self, path: Path) -> None:
        """Set the output path for testing.

        Args:
            path: path to write telemetry.

        """
        os.makedirs(path, exist_ok=True)
        self.sc_folder_path = path

    def sc_log_error(
        self, event_name: str, parameters_info: Optional[dict] = None
    ) -> None:
        """Log an error telemetry event.

        Args:
            event_name (str): The name of the event.
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        """
        self._sc_log_telemetry(event_name, "error", parameters_info)

    def sc_log_info(
        self, event_name: str, parameters_info: Optional[dict] = None
    ) -> None:
        """Log an information telemetry event.

        Args:
            event_name (str): The name of the event.
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        """
        self._sc_log_telemetry(event_name, "info", parameters_info)

    def _sc_log_telemetry(
        self, event_name: str, event_type: str, parameters_info: Optional[dict] = None
    ) -> dict:
        """Log a telemetry event if is enabled.

        Args:
            event_name (str): The name of the event.
            event_type (str): The type of the event (e.g., "error", "info").
            parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

        Returns:
            dict: The logged event.

        """
        if not self.sc_is_enabled:
            return {}
        event = _generate_event(event_name, event_type, parameters_info)
        self._sc_add_log_to_batch(event)
        return event

    def _sc_add_log_to_batch(self, event: dict) -> None:
        """Add a log event to the batch. If the batch is full, send the events to the API.

        Args:
            event (dict): The event to add.

        """
        self.sc_log_batch.append(event)
        if self.sc_is_testing:
            self._sc_write_telemetry(self.sc_log_batch)
            self.sc_log_batch = []
            return

        if len(self.sc_log_batch) >= self.sc_flush_size:
            self._sc_send_batch(self.sc_log_batch)
            self.sc_log_batch = []

    def _sc_send_batch(self, to_sent: list) -> bool:
        """Send a request to the API to upload the events. If not have connection, write the events to local folder.

        Args:
            to_sent (list): The batch of events to send.

        Returns:
            bool: True if the batch was sent successfully, False otherwise.

        """
        if not self.sc_is_enabled:
            return False
        if self._rest is None:
            self._sc_write_telemetry(to_sent)
            self.sc_log_batch = []
            return False
        body = {"logs": to_sent}
        ret = self._rest.request(
            self.sc_sf_path_telemetry,
            body=body,
            method="post",
            client=None,
            timeout=5,
        )
        if not ret.get("success"):
            self._sc_write_telemetry(to_sent)
            self.sc_log_batch = []
            return False
        return True

    def _sc_write_telemetry(self, batch: list) -> None:
        """Write telemetry events to local folder. If the folder is full, free up space for the new events.

        Args:
            batch (list): The batch of events to write.

        """
        try:
            makedirs(self.sc_folder_path, exist_ok=True)
            for event in batch:
                message = event.get("message")
                if message is not None:
                    file_path = (
                        self.sc_folder_path
                        / f'{datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")}'
                        f'_telemetry_{message.get("type")}.json'
                    )
                    json_content = self._sc_validate_folder_space(event)
                    with open(file_path, "w") as json_file:
                        json_file.write(json_content)
        except Exception:
            pass

    def _sc_validate_folder_space(self, event: dict) -> str:
        """Validate and manage folder space for the new telemetry events.

        Args:
            event (dict): The event to validate.

        Returns:
            str: The JSON content of the event.

        """
        json_content = json.dumps(event, indent=4, sort_keys=True)
        new_json_file_size = len(json_content.encode("utf-8"))
        telemetry_folder = self.sc_folder_path
        folder_size = _get_folder_size(telemetry_folder)
        if folder_size + new_json_file_size > self.sc_memory_limit:
            _free_up_space(telemetry_folder, self.sc_memory_limit - new_json_file_size)
        return json_content

    def _sc_upload_local_telemetry(self) -> None:
        """Send a request to the API to upload the local telemetry events."""
        if not self.is_enabled() or self.sc_is_testing:
            return
        batch = []
        for file in self.sc_folder_path.glob("*.json"):
            with open(file) as json_file:
                data_dict = json.load(json_file)
                batch.append(data_dict)
        body = {"logs": batch}
        ret = self._rest.request(
            self.sc_sf_path_telemetry,
            body=body,
            method="post",
            client=None,
            timeout=5,
        )
        if ret.get("success"):
            for file in self.sc_folder_path.glob("*.json"):
                file.unlink()

    def _sc_is_telemetry_enabled(self) -> bool:
        """Check if telemetry is enabled.

        Returns:
            bool: True if telemetry is enabled, False otherwise.

        """
        if self._sc_is_telemetry_testing():
            return True
        if getenv("SNOWPARK_CHECKPOINTS_TELEMETRY_ENABLED") == "false":
            return False
        return self._rest is not None

    def _sc_is_telemetry_testing(self) -> bool:
        is_testing = getenv("SNOWPARK_CHECKPOINTS_TELEMETRY_TESTING") == "true"
        if is_testing:
            local_telemetry_path = (
                Path(getcwd()) / "snowpark-checkpoints-output" / "telemetry"
            )
            self.set_sc_output_path(local_telemetry_path)
        return is_testing

    def _sc_is_telemetry_manager(self) -> bool:
        """Check if the class is telemetry manager.

        Returns:
            bool: True if the class is telemetry manager, False otherwise.

        """
        return True

    def sc_is_hypothesis_event_logged(self, event_name: tuple[str, int]) -> bool:
        """Check if a Hypothesis event is logged.

        Args:
            event_name (tuple[str, int]): A tuple containing the name of the event and an integer identifier
            (0 for info, 1 for error).

        Returns:
            bool: True if the event is logged, False otherwise.

        """
        return event_name in self.sc_hypothesis_input_events


def _generate_event(
    event_name: str,
    event_type: str,
    parameters_info: Optional[dict] = None,
) -> dict:
    """Generate a telemetry event.

    Args:
        event_name (str): The name of the event.
        event_type (str): The type of the event (e.g., "error", "info").
        parameters_info (dict, optional): Additional parameters for the event. Defaults to None.

    Returns:
        dict: The generated event.

    """
    metadata = _get_metadata()
    message = {
        "type": event_type,
        "event_name": event_name,
        "driver_type": "PythonConnector",
        "driver_version": SNOWFLAKE_CONNECTOR_VERSION,
        "source": "snowpark-checkpoints",
        "metadata": metadata,
        "data": json.dumps(parameters_info or {}),
    }
    timestamp = time_util.get_time_millis()
    event_base = {"message": message, "timestamp": str(timestamp)}

    return event_base


def _get_metadata() -> dict:
    """Get metadata for telemetry events.

    Returns:
        dict: The metadata including OS version, Python version, and device ID.

    """
    return {
        "os_version": platform,
        "python_version": python_version(),
        "snowpark_version": ".".join(str(x) for x in SNOWPARK_VERSION if x is not None),
        "device_id": _get_unique_id(),
    }


def _get_folder_size(folder_path: Path) -> int:
    """Get the size of a folder. Only considers JSON files.

    Args:
        folder_path (Path): The path to the folder.

    Returns:
        int: The size of the folder in bytes.

    """
    return sum(f.stat().st_size for f in folder_path.glob("*.json") if f.is_file())


def _free_up_space(folder_path: Path, max_size: int) -> None:
    """Free up space in a folder by deleting the oldest files. Only considers JSON files.

    Args:
        folder_path (Path): The path to the folder.
        max_size (int): The maximum allowed size of the folder in bytes.

    """
    files = sorted(folder_path.glob("*.json"), key=lambda f: f.stat().st_mtime)
    current_size = _get_folder_size(folder_path)
    for file in files:
        if current_size <= max_size:
            break
        current_size -= file.stat().st_size
        file.unlink()


def _get_unique_id() -> str:
    """Get a unique device ID. The ID is generated based on the hashed MAC address.

    Returns:
        str: The hashed device ID.

    """
    node_id_str = str(getnode())
    hashed_id = hashlib.sha256(node_id_str.encode()).hexdigest()
    return hashed_id


def get_telemetry_manager() -> TelemetryManager:
    """Get the telemetry manager.

    Returns:
        TelemetryManager: The telemetry manager.

    """
    connection = Session.builder.getOrCreate().connection
    if not hasattr(connection._telemetry, "_sc_is_telemetry_manager"):
        connection._telemetry = TelemetryManager(connection._rest)
    return connection._telemetry


def get_snowflake_schema_types(df: snowpark_dataframe.DataFrame) -> list[str]:
    """Extract the data types of the schema fields from a Snowflake DataFrame.

    Args:
        df (snowpark_dataframe.DataFrame): The Snowflake DataFrame.

    Returns:
        list[str]: A list of data type names of the schema fields.

    """
    return [str(schema_type.datatype) for schema_type in df.schema.fields]


def get_load_json(json_schema: str) -> dict:
    """Load and parse a JSON schema file.

    Args:
        json_schema (str): The path to the JSON schema file.

    Returns:
        dict: The parsed JSON content.

    Raises:
        ValueError: If there is an error reading or parsing the JSON file.

    """
    try:
        with open(json_schema, encoding="utf-8") as file:
            return json.load(file)
    except (OSError, json.JSONDecodeError) as e:
        raise ValueError(f"Error reading JSON schema file: {e}") from None


def extract_parameters(
    func: Callable, args: tuple, kwargs: dict, params_list: Optional[list[str]]
) -> dict:
    """Extract parameters from the function arguments.

    Args:
        func (Callable): The function being decorated.
        args (tuple): The positional arguments passed to the function.
        kwargs (dict): The keyword arguments passed to the function.
        params_list (list[str]): The list of parameters to extract.

    Returns:
        dict: A dictionary of extracted parameters.

    """
    parameters = inspect.signature(func).parameters
    param_data = {}
    if params_list:
        for _, param in enumerate(params_list):
            if len(args) > 0:
                index = list(parameters.keys()).index(param)
                param_data[param] = args[index]
            else:
                if kwargs[param]:
                    param_data[param] = kwargs[param]
    return param_data


def handle_result(
    func_name: str,
    result: Any,
    param_data: dict,
    return_indexes: Optional[list[tuple[str, int]]],
    multiple_return: bool,
    telemetry_m: TelemetryManager,
) -> None:
    """Handle the result of the function and log telemetry data.

    Args:
        func_name (str): The name of the function.
        result: The result of the function.
        param_data (dict): The extracted parameters.
        return_indexes (list[tuple[str, int]]): The list of return values to report.
        multiple_return (bool): Whether the function returns multiple values.
        telemetry_m (TelemetryManager): The telemetry manager.

    """
    if result and return_indexes:
        if multiple_return:
            for name, index in return_indexes:
                param_data[name] = result[index]
        else:
            param_data[return_indexes[0][0]] = result[return_indexes[0][1]]

    telemetry_data = {
        FUNCTION_KEY: func_name,
    }

    if func_name == "_check_dataframe_schema":
        telemetry_data[STATUS_KEY] = param_data[STATUS_KEY]
        telemetry_data[SCHEMA_TYPES_KEY] = list(
            param_data.get("pandera_schema").columns.keys()
        )
        telemetry_m.sc_log_info(DATAFRAME_VALIDATOR_SCHEMA, telemetry_data)
    elif func_name in ["check_output_schema", "check_input_schema"]:
        telemetry_data[SCHEMA_TYPES_KEY] = list(
            param_data.get("pandera_schema").columns.keys()
        )
        telemetry_m.sc_log_info(DATAFRAME_VALIDATOR_SCHEMA, telemetry_data)
    elif func_name == "_collect_dataframe_checkpoint_mode_schema":
        telemetry_data[MODE_KEY] = CheckpointMode.SCHEMA.value
        schema_types = param_data.get("column_type_dict")
        telemetry_data[SCHEMA_TYPES_KEY] = [
            schema_types[schema_type].dataType.typeName()
            for schema_type in schema_types
        ]
        telemetry_m.sc_log_info(DATAFRAME_COLLECTION, telemetry_data)
    elif func_name == "_assert_return":
        telemetry_data[STATUS_KEY] = param_data[STATUS_KEY]
        if isinstance(
            param_data.get("snowpark_results"), snowpark_dataframe.DataFrame
        ) and _is_spark_dataframe(param_data.get("spark_results")):
            telemetry_data[SNOWFLAKE_SCHEMA_TYPES_KEY] = get_snowflake_schema_types(
                param_data.get("snowpark_results")
            )
            telemetry_data[SPARK_SCHEMA_TYPES_KEY] = _get_spark_schema_types(
                param_data.get("spark_results")
            )
            telemetry_m.sc_log_info(DATAFRAME_VALIDATOR_MIRROR, telemetry_data)
        else:
            telemetry_m.sc_log_info(VALUE_VALIDATOR_MIRROR, telemetry_data)
    elif func_name == "dataframe_strategy":
        schema_param = param_data.get(DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME)
        if isinstance(schema_param, str):
            is_logged = telemetry_m.sc_is_hypothesis_event_logged((schema_param, 0))
            if not is_logged:
                json_data = get_load_json(schema_param)["custom_data"]["columns"]
                telemetry_data[SCHEMA_TYPES_KEY] = [
                    column["type"] for column in json_data
                ]
                telemetry_m.sc_log_info(HYPOTHESIS_INPUT_SCHEMA, telemetry_data)
                telemetry_m.sc_hypothesis_input_events.append((schema_param, 0))


def handle_exception(func_name: str, param_data: dict, err: Exception) -> None:
    """Handle exceptions raised by the function and log telemetry data.

    Args:
        func_name (str): The name of the function.
        param_data (dict): The extracted parameters.
        err (Exception): The exception raised by the function.

    """
    telemetry_m = get_telemetry_manager()
    telemetry_data = {
        FUNCTION_KEY: func_name,
        ERROR_KEY: type(err).__name__,
    }
    if func_name == "_collect_dataframe_checkpoint_mode_schema":
        schema_types = param_data.get("column_type_dict")
        if schema_types:
            telemetry_data[SCHEMA_TYPES_KEY] = [
                schema_types[schema_type].dataType.typeName()
                for schema_type in schema_types
            ]
        telemetry_m.sc_log_error(DATAFRAME_COLLECTION_ERROR, telemetry_data)
    elif func_name in [
        "_check_dataframe_schema",
        "check_output_schema",
        "check_input_schema",
    ]:
        pandera_schema = param_data.get("pandera_schema")
        if pandera_schema:
            schema_types = pandera_schema.columns.keys()
            telemetry_data[SCHEMA_TYPES_KEY] = list(schema_types)
        telemetry_m.sc_log_error(DATAFRAME_VALIDATOR_ERROR, telemetry_data)
    elif func_name == "_assert_return":
        if isinstance(param_data.get("snowpark_results"), snowpark_dataframe.DataFrame):
            telemetry_data[SNOWFLAKE_SCHEMA_TYPES_KEY] = get_snowflake_schema_types(
                param_data.get("snowpark_results")
            )
        if _is_spark_dataframe(param_data.get("spark_results")):
            telemetry_data[SPARK_SCHEMA_TYPES_KEY] = _get_spark_schema_types(
                param_data.get("spark_results")
            )
        telemetry_m.sc_log_error(DATAFRAME_VALIDATOR_ERROR, telemetry_data)
    elif func_name == "dataframe_strategy":
        input_json = param_data.get(DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME)
        if input_json:
            is_logged = telemetry_m.sc_is_hypothesis_event_logged((input_json, 1))
            if not is_logged:
                telemetry_m.sc_log_error(HYPOTHESIS_INPUT_SCHEMA_ERROR, telemetry_data)
                if isinstance(input_json, str):
                    telemetry_m.sc_hypothesis_input_events.append(
                        (param_data.get(DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME), 1)
                    )


fn = TypeVar("fn", bound=Callable)


def report_telemetry(
    params_list: list[str] = None,
    return_indexes: list[tuple[str, int]] = None,
    multiple_return: bool = False,
) -> Callable[[fn], fn]:
    """Report telemetry events for a function.

    Args:
        params_list (list[str], optional): The list of parameters to report. Defaults to None.
        return_indexes (list[tuple[str, int]], optional): The list of return values to report. Defaults to None.
        multiple_return (bool, optional): Whether the function returns multiple values. Defaults to False.

    Returns:
        Callable[[fn], fn]: The decorator function.

    """

    def report_telemetry_decorator(func):
        func_name = func.__name__

        @wraps(func)
        def wrapper(*args, **kwargs):
            func_exception = None
            result = None
            try:
                result = func(*args, **kwargs)
            except Exception as err:
                func_exception = err

            param_data = {}
            try:
                param_data = extract_parameters(func, args, kwargs, params_list)
                telemetry_m = get_telemetry_manager()
                handle_result(
                    func_name,
                    result,
                    param_data,
                    return_indexes,
                    multiple_return,
                    telemetry_m,
                )
            except Exception as err:
                handle_exception(func_name, param_data, err)

            if func_exception is not None:
                raise func_exception
            return result

        return wrapper

    return report_telemetry_decorator


# Constants for telemetry
DATAFRAME_COLLECTION = "DataFrame_Collection"
DATAFRAME_VALIDATOR_MIRROR = "DataFrame_Validator_Mirror"
VALUE_VALIDATOR_MIRROR = "Value_Validator_Mirror"
DATAFRAME_VALIDATOR_SCHEMA = "DataFrame_Validator_Schema"
HYPOTHESIS_INPUT_SCHEMA = "Hypothesis_Input_Schema"
DATAFRAME_COLLECTION_ERROR = "DataFrame_Collection_Error"
DATAFRAME_VALIDATOR_ERROR = "DataFrame_Validator_Error"
HYPOTHESIS_INPUT_SCHEMA_ERROR = "Hypothesis_Input_Schema_Error"

FUNCTION_KEY = "function"
STATUS_KEY = "status"
SCHEMA_TYPES_KEY = "schema_types"
ERROR_KEY = "error"
MODE_KEY = "mode"
SNOWFLAKE_SCHEMA_TYPES_KEY = "snowflake_schema_types"
SPARK_SCHEMA_TYPES_KEY = "spark_schema_types"

DATAFRAME_STRATEGY_SCHEMA_PARAM_NAME = "schema"


class CheckpointMode(IntEnum):
    SCHEMA = 1
    DATAFRAME = 2



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

import locale

from enum import IntEnum


class CheckpointMode(IntEnum):
    SCHEMA = 1
    DATAFRAME = 2


# CONSTANTS
ARRAY_COLUMN_TYPE = "array"
BINARY_COLUMN_TYPE = "binary"
BOOLEAN_COLUMN_TYPE = "boolean"
BYTE_COLUMN_TYPE = "byte"
DATE_COLUMN_TYPE = "date"
DAYTIMEINTERVAL_COLUMN_TYPE = "daytimeinterval"
DECIMAL_COLUMN_TYPE = "decimal"
DOUBLE_COLUMN_TYPE = "double"
FLOAT_COLUMN_TYPE = "float"
INTEGER_COLUMN_TYPE = "integer"
LONG_COLUMN_TYPE = "long"
MAP_COLUMN_TYPE = "map"
NULL_COLUMN_TYPE = "void"
SHORT_COLUMN_TYPE = "short"
STRING_COLUMN_TYPE = "string"
STRUCT_COLUMN_TYPE = "struct"
TIMESTAMP_COLUMN_TYPE = "timestamp"
TIMESTAMP_NTZ_COLUMN_TYPE = "timestamp_ntz"

PANDAS_BOOLEAN_DTYPE = "bool"
PANDAS_DATETIME_DTYPE = "datetime64[ns]"
PANDAS_FLOAT_DTYPE = "float64"
PANDAS_INTEGER_DTYPE = "int64"
PANDAS_OBJECT_DTYPE = "object"
PANDAS_TIMEDELTA_DTYPE = "timedelta64[ns]"

NUMERIC_TYPE_COLLECTION = [
    BYTE_COLUMN_TYPE,
    DOUBLE_COLUMN_TYPE,
    FLOAT_COLUMN_TYPE,
    INTEGER_COLUMN_TYPE,
    LONG_COLUMN_TYPE,
    SHORT_COLUMN_TYPE,
]

INTEGER_TYPE_COLLECTION = [
    BYTE_COLUMN_TYPE,
    INTEGER_COLUMN_TYPE,
    LONG_COLUMN_TYPE,
    SHORT_COLUMN_TYPE,
]

PANDAS_OBJECT_TYPE_COLLECTION = [
    STRING_COLUMN_TYPE,
    ARRAY_COLUMN_TYPE,
    MAP_COLUMN_TYPE,
    NULL_COLUMN_TYPE,
    STRUCT_COLUMN_TYPE,
]

BETWEEN_CHECK_ERROR_MESSAGE_FORMAT = "Value must be between {} and {}"

# SCHEMA CONTRACT KEYS CONSTANTS
COLUMN_ALLOW_NULL_KEY = "allow_null"
COLUMN_COUNT_KEY = "rows_count"
COLUMN_DECIMAL_PRECISION_KEY = "decimal_precision"
COLUMN_FALSE_COUNT_KEY = "false_count"
COLUMN_FORMAT_KEY = "format"
COLUMN_IS_NULLABLE_KEY = "nullable"
COLUMN_IS_UNIQUE_SIZE_KEY = "is_unique_size"
COLUMN_KEY_TYPE_KEY = "key_type"
COLUMN_MARGIN_ERROR_KEY = "margin_error"
COLUMN_MAX_KEY = "max"
COLUMN_MAX_SIZE_KEY = "max_size"
COLUMN_MEAN_KEY = "mean"
COLUMN_MEAN_SIZE_KEY = "mean_size"
COLUMN_METADATA_KEY = "metadata"
COLUMN_MIN_KEY = "min"
COLUMN_MIN_SIZE_KEY = "min_size"
COLUMN_NAME_KEY = "name"
COLUMN_NULL_COUNT_KEY = "null_count"
COLUMN_NULL_VALUE_PROPORTION_KEY = "null_value_proportion"
COLUMN_ROWS_NOT_NULL_COUNT_KEY = "rows_not_null_count"
COLUMN_ROWS_NULL_COUNT_KEY = "rows_null_count"
COLUMN_SIZE_KEY = "size"
COLUMN_TRUE_COUNT_KEY = "true_count"
COLUMN_TYPE_KEY = "type"
COLUMN_VALUE_TYPE_KEY = "value_type"
COLUMNS_KEY = "columns"

DATAFRAME_CUSTOM_DATA_KEY = "custom_data"
DATAFRAME_PANDERA_SCHEMA_KEY = "pandera_schema"

PANDERA_COLUMN_TYPE_KEY = "dtype"

CONTAINS_NULL_KEY = "containsNull"
ELEMENT_TYPE_KEY = "elementType"
FIELD_METADATA_KEY = "metadata"
FIELDS_KEY = "fields"
KEY_TYPE_KEY = "keyType"
NAME_KEY = "name"
VALUE_CONTAINS_NULL_KEY = "valueContainsNull"
VALUE_TYPE_KEY = "valueType"

# DIRECTORY AND FILE NAME CONSTANTS
SNOWPARK_CHECKPOINTS_OUTPUT_DIRECTORY_NAME = "snowpark-checkpoints-output"
CHECKPOINT_JSON_OUTPUT_FILE_NAME_FORMAT = "{}.json"
CHECKPOINT_PARQUET_OUTPUT_FILE_NAME_FORMAT = "{}.parquet"
COLLECTION_RESULT_FILE_NAME = "checkpoint_collection_results.json"

# MISC KEYS
DECIMAL_TOKEN_KEY = "decimal_point"
DOT_PARQUET_EXTENSION = ".parquet"
DOT_IPYNB_EXTENSION = ".ipynb"
UNKNOWN_SOURCE_FILE = "unknown"
UNKNOWN_LINE_OF_CODE = -1
BACKSLASH_TOKEN = "\\"
SLASH_TOKEN = "/"

# ENVIRONMENT VARIABLES
SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH_ENV_VAR = (
    "SNOWFLAKE_CHECKPOINT_CONTRACT_FILE_PATH"
)


def get_decimal_token() -> str:
    """Return the decimal token based on the local environment.

    Returns:
        str: The decimal token.

    """
    decimal_token = locale.localeconv()[DECIMAL_TOKEN_KEY]
    return decimal_token



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from abc import ABC, abstractmethod

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_COUNT_KEY,
    COLUMN_IS_NULLABLE_KEY,
    COLUMN_NAME_KEY,
    COLUMN_ROWS_NOT_NULL_COUNT_KEY,
    COLUMN_ROWS_NULL_COUNT_KEY,
    COLUMN_TYPE_KEY,
)


class ColumnCollectorBase(ABC):

    """Base class for column collector based on type.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init ColumnCollectorBase.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        self.name = clm_name
        self.type = struct_field.dataType.typeName()
        self.struct_field = struct_field
        self.values = clm_values

    @abstractmethod
    def get_custom_data(self) -> dict[str, any]:
        """Get the custom data of the column.

        Returns:
            dict[str, any]: The data collected.

        """
        pass

    def _get_common_data(self) -> dict[str, any]:
        column_size = len(self.values)
        rows_not_null_count = self.values.count().item()
        rows_null_count = column_size - rows_not_null_count

        common_data_dict = {
            COLUMN_NAME_KEY: self.name,
            COLUMN_TYPE_KEY: self.type,
            COLUMN_IS_NULLABLE_KEY: self.struct_field.nullable,
            COLUMN_COUNT_KEY: column_size,
            COLUMN_ROWS_NOT_NULL_COUNT_KEY: rows_not_null_count,
            COLUMN_ROWS_NULL_COUNT_KEY: rows_null_count,
        }

        return common_data_dict

    def get_data(self) -> dict[str, any]:
        """Get the data collected of the column.

        Returns:
            dict[str, any]: The data collected.

        """
        common_data = self._get_common_data()
        custom_data = self.get_custom_data()
        column_data = dict(common_data | custom_data)
        return column_data



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from datetime import date, timedelta
from typing import Optional

import hypothesis.strategies as st
import pandera as pa

from pandas import Series
from pandera import extensions


def date_strategy(
    pandera_dtype: pa.DataType,
    strategy: Optional[st.SearchStrategy] = None,
    *,
    min_value: date,
    max_value: date,
    include_min: bool,
    include_max: bool,
):
    """Create a Hypothesis strategy for generating dates within a specified range.

    Args:
        pandera_dtype: The Pandera data type.
        strategy: The chained strategy to use.
        min_value: The minimum date value.
        max_value: The maximum date value.
        include_min: Whether to include the minimum date value.
        include_max: Whether to include the maximum date value.

    Returns:
        A Hypothesis strategy that generates dates within the specified range.

    """
    if strategy is not None:
        raise ValueError("Chaining strategies is not supported for 'date' dtype.")

    if not include_min:
        min_value = min_value + timedelta(days=1)
    if not include_max:
        max_value = max_value - timedelta(days=1)

    if min_value > max_value:
        raise ValueError(
            "Invalid range: min_value must be less than or equal to max_value."
        )

    return st.dates(min_value=min_value, max_value=max_value)


@extensions.register_check_method(
    statistics=["min_value", "max_value", "include_min", "include_max"],
    strategy=date_strategy,
)
def dates_in_range(
    pandas_obj: Series,
    *,
    min_value: date,
    max_value: date,
    include_min: True,
    include_max: True,
) -> Series:
    """Check if a date is within the specified range.

    Args:
        pandas_obj: The pandas Series object to check.
        min_value: The minimum date value.
        max_value: The maximum date value.
        include_min: Whether to include the minimum date value.
        include_max: Whether to include the maximum date value.

    Returns:
        True if the date is within the specified range, False otherwise.

    """
    actual_date = pandas_obj.dt.date if hasattr(pandas_obj, "dt") else pandas_obj
    is_in_range = (
        actual_date >= min_value if include_min else actual_date > min_value
    ) & (actual_date <= max_value if include_max else actual_date < max_value)
    return is_in_range



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = ["collect_dataframe_checkpoint", "Singleton"]

from snowflake.snowpark_checkpoints_collector.singleton import Singleton
from snowflake.snowpark_checkpoints_collector.summary_stats_collector import (
    collect_dataframe_checkpoint,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import os

from snowflake.snowpark_checkpoints_configuration.model.checkpoints import (
    Checkpoint,
    Checkpoints,
)
from snowflake.snowpark_checkpoints_configuration.singleton import Singleton


class CheckpointMetadata(metaclass=Singleton):

    """CheckpointMetadata class.

     This is a singleton class that reads the checkpoints.json file
    and provides an interface to get the checkpoint configuration.

    Args:
        metaclass (Singleton, optional): Defaults to Singleton.

    """

    def __init__(self, path: str = None):
        directory = path if path is not None else os.getcwd()
        self.checkpoint_model: Checkpoints = Checkpoints(type="", pipelines=[])
        checkpoints_file = os.path.join(directory, "checkpoints.json")
        if os.path.exists(checkpoints_file):
            with open(checkpoints_file) as f:
                try:
                    checkpoint_json = f.read()
                    self.checkpoint_model = Checkpoints.model_validate_json(
                        checkpoint_json
                    )
                except Exception as e:
                    raise Exception(
                        f"Error reading checkpoints file: {checkpoints_file} \n {e}"
                    ) from None

    def get_checkpoint(self, checkpoint_name: str) -> Checkpoint:
        """Get a checkpoint by its name.

        Args:
            checkpoint_name (str): checkpoint name

        Returns:
            Checkpoint: Checkpoint configuration instance

        """
        return self.checkpoint_model.get_check_point(checkpoint_name=checkpoint_name)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = ["CollectionPointResult", "CollectionResult", "CollectionPointResultManager"]

from snowflake.snowpark_checkpoints_collector.collection_result.model.collection_point_result import (
    CollectionPointResult,
    CollectionResult,
)
from snowflake.snowpark_checkpoints_collector.collection_result.model.collection_point_result_manager import (
    CollectionPointResultManager,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

__all__ = ["PanderaColumnChecksManager"]

from snowflake.snowpark_checkpoints_collector.column_pandera_checks.pandera_column_checks_manager import (
    PanderaColumnChecksManager,
)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from pandas import Series
from pyspark.sql.types import StructField

from snowflake.snowpark_checkpoints_collector.collection_common import (
    COLUMN_MAX_KEY,
    COLUMN_MIN_KEY,
)
from snowflake.snowpark_checkpoints_collector.column_collection.model.column_collector_base import (
    ColumnCollectorBase,
)


class DayTimeIntervalColumnCollector(ColumnCollectorBase):

    """Class for collect a date time interval type column.

    Attributes:
        name (str): the name of the column.
        type (str): the type of the column.
        struct_field (pyspark.sql.types.StructField): the struct field of the column type.
        values (pandas.Series): the column values as Pandas.Series.

    """

    def __init__(
        self, clm_name: str, struct_field: StructField, clm_values: Series
    ) -> None:
        """Init DayTimeIntervalColumnCollector.

        Args:
            clm_name (str): the name of the column.
            struct_field (pyspark.sql.types.StructField): the struct field of the column type.
            clm_values (pandas.Series): the column values as Pandas.Series.

        """
        super().__init__(clm_name, struct_field, clm_values)

    def get_custom_data(self) -> dict[str, any]:
        local_values = self.values.dropna()
        min_value = str(local_values.min())
        max_value = str(local_values.max())

        custom_data_dict = {COLUMN_MIN_KEY: min_value, COLUMN_MAX_KEY: max_value}

        return custom_data_dict



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from pandas import DataFrame as PandasDataFrame
from pandera import Check, Column

from snowflake.snowpark_checkpoints_collector.collection_common import (
    BETWEEN_CHECK_ERROR_MESSAGE_FORMAT,
    BOOLEAN_COLUMN_TYPE,
    BYTE_COLUMN_TYPE,
    DATE_COLUMN_TYPE,
    DAYTIMEINTERVAL_COLUMN_TYPE,
    DOUBLE_COLUMN_TYPE,
    FLOAT_COLUMN_TYPE,
    INTEGER_COLUMN_TYPE,
    LONG_COLUMN_TYPE,
    SHORT_COLUMN_TYPE,
    STRING_COLUMN_TYPE,
    TIMESTAMP_COLUMN_TYPE,
    TIMESTAMP_NTZ_COLUMN_TYPE,
)


def collector_register(cls):
    """Decorate a class with the checks mechanism.

    Args:
        cls: The class to decorate.

    Returns:
        The class to decorate.

    """
    cls._collectors = {}
    for method_name in dir(cls):
        method = getattr(cls, method_name)
        if hasattr(method, "_column_type"):
            col_type_collection = method._column_type
            for col_type in col_type_collection:
                cls._collectors[col_type] = method_name
    return cls


def column_register(*args):
    """Decorate a method to register it in the checks mechanism based on column type.

    Args:
        args: the column type to register.

    Returns:
        The wrapper.

    """

    def wrapper(func):
        has_arguments = len(args) > 0
        if has_arguments:
            func._column_type = args
        return func

    return wrapper


def _datetime_like_type_checks(
    clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
) -> None:
    column_values = pandas_df[clm_name].dropna()
    min_value = str(column_values.min())
    max_value = str(column_values.max())
    pandera_column.checks.append(
        Check.between(
            min_value=min_value,
            max_value=max_value,
            include_max=True,
            include_min=True,
            title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
        )
    )


@collector_register
class PanderaColumnChecksManager:

    """Manage class for Pandera column checks based on type."""

    def add_checks_column(
        self,
        clm_name: str,
        clm_type: str,
        pandas_df: PandasDataFrame,
        pandera_column: Column,
    ) -> None:
        """Add checks to Pandera column based on the column type.

        Args:
            clm_name (str): the name of the column.
            clm_type (str): the type of the column.
            pandas_df (pandas.DataFrame): the DataFrame.
            pandera_column (pandera.Column): the Pandera column.

        """
        if clm_type not in self._collectors:
            return

        func_name = self._collectors[clm_type]
        func = getattr(self, func_name)
        func(clm_name, pandas_df, pandera_column)

    @column_register(BOOLEAN_COLUMN_TYPE)
    def _add_boolean_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        pandera_column.checks.extend([Check.isin([True, False])])

    @column_register(DATE_COLUMN_TYPE)
    def _add_date_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        _datetime_like_type_checks(clm_name, pandas_df, pandera_column)

    @column_register(DAYTIMEINTERVAL_COLUMN_TYPE)
    def _add_daytimeinterval_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        _datetime_like_type_checks(clm_name, pandas_df, pandera_column)

    @column_register(
        BYTE_COLUMN_TYPE,
        SHORT_COLUMN_TYPE,
        INTEGER_COLUMN_TYPE,
        LONG_COLUMN_TYPE,
        FLOAT_COLUMN_TYPE,
        DOUBLE_COLUMN_TYPE,
    )
    def _add_numeric_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        column_values = pandas_df[clm_name].dropna()
        min_value = column_values.min().item()
        max_value = column_values.max().item()
        pandera_column.checks.append(
            Check.between(
                min_value=min_value,
                max_value=max_value,
                include_max=True,
                include_min=True,
                title=BETWEEN_CHECK_ERROR_MESSAGE_FORMAT.format(min_value, max_value),
            )
        )

    @column_register(STRING_COLUMN_TYPE)
    def _add_string_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        pass

    @column_register(TIMESTAMP_COLUMN_TYPE)
    def _add_timestamp_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        _datetime_like_type_checks(clm_name, pandas_df, pandera_column)

    @column_register(TIMESTAMP_NTZ_COLUMN_TYPE)
    def _add_timestamp_ntz_type_checks(
        self, clm_name: str, pandas_df: PandasDataFrame, pandera_column: Column
    ) -> None:
        _datetime_like_type_checks(clm_name, pandas_df, pandera_column)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
from datetime import datetime
from enum import Enum

from snowflake.snowpark_checkpoints_collector.utils import file_utils


TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S"

TIMESTAMP_KEY = "timestamp"
FILE_KEY = "file"
RESULT_KEY = "result"
LINE_OF_CODE_KEY = "line_of_code"
CHECKPOINT_NAME_KEY = "checkpoint_name"


class CollectionResult(Enum):
    FAIL = "FAIL"
    PASS = "PASS"


class CollectionPointResult:

    """Class for checkpoint collection results.

    Attributes:
        _timestamp (timestamp): the timestamp when collection started.
        _file_path (str): the full path where checkpoint is.
        _line_of_code (int): the line of code where the checkpoint is.
        _checkpoint_name (str): the checkpoint name.
        _result (CollectionResult): the result status of the checkpoint collection point.

    """

    def __init__(
        self,
        file_path: str,
        line_of_code: int,
        checkpoint_name: str,
    ) -> None:
        """Init CollectionPointResult.

        Args:
            file_path (str): the full path where checkpoint is.
            line_of_code (int): the line of code where the checkpoint is.
            checkpoint_name (str): the checkpoint name.

        """
        self._timestamp = datetime.now()
        self._file_path = file_path
        self._line_of_code = line_of_code
        self._checkpoint_name = checkpoint_name
        self._result = None

    @property
    def result(self):
        """Get the result status of the checkpoint collection point."""
        return self._result

    @result.setter
    def result(self, value):
        """Set the result status of the checkpoint collection point."""
        self._result = value

    def get_collection_result_data(self) -> dict[str, any]:
        """Get the results of the checkpoint collection point."""
        timestamp_with_format = self._timestamp.strftime(TIMESTAMP_FORMAT)
        relative_path = file_utils.get_relative_file_path(self._file_path)

        collection_point_result = {
            TIMESTAMP_KEY: timestamp_with_format,
            FILE_KEY: relative_path,
            LINE_OF_CODE_KEY: self._line_of_code,
            CHECKPOINT_NAME_KEY: self._checkpoint_name,
            RESULT_KEY: self.result.value,
        }

        return collection_point_result



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#
import json

from typing import Optional

from snowflake.snowpark_checkpoints_collector import Singleton
from snowflake.snowpark_checkpoints_collector.collection_result.model import (
    CollectionPointResult,
)
from snowflake.snowpark_checkpoints_collector.utils import file_utils


RESULTS_KEY = "results"


class CollectionPointResultManager(metaclass=Singleton):

    """Class for manage the checkpoint collection results. It is a singleton.

    Attributes:
        result_collection (list[any]): the collection of the checkpoint results.
        output_file_path (str): the full path of the output file.

    """

    def __init__(self, output_path: Optional[str] = None) -> None:
        """Init CollectionPointResultManager."""
        self.result_collection: list[any] = []
        self.output_file_path = file_utils.get_output_file_path(output_path)

    def add_result(self, result: CollectionPointResult) -> None:
        """Add the CollectionPointResult result to the collection.

        Args:
            result (CollectionPointResult): the CollectionPointResult to add.

        """
        result_json = result.get_collection_result_data()
        self.result_collection.append(result_json)
        self._save_result()

    def to_json(self) -> str:
        """Convert to json the checkpoint results collected.

        Returns:
            str: the results as json string.

        """
        dict_object = {RESULTS_KEY: self.result_collection}
        result_collection_json = json.dumps(dict_object)
        return result_collection_json

    def _save_result(self) -> None:
        result_collection_json = self.to_json()
        with open(self.output_file_path, "w") as f:
            f.write(result_collection_json)



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from datetime import datetime
from typing import Optional

import pandas as pd

from pyspark.sql import SparkSession

from snowflake.snowpark import Session
from snowflake.snowpark_checkpoints.utils.constant import SCHEMA_EXECUTION_MODE


class SnowparkJobContext:
    def __init__(
        self,
        snowpark_session: Session,
        spark_session: SparkSession = None,
        job_name: Optional[str] = None,
        log_results: Optional[bool] = True,
    ):
        self.log_results = log_results
        self.job_name = job_name
        self.spark_session = spark_session or SparkSession.builder.getOrCreate()
        self.snowpark_session = snowpark_session

    def mark_fail(
        self, message, checkpoint_name, data, execution_mode=SCHEMA_EXECUTION_MODE
    ):
        if self.log_results:
            session = self.snowpark_session
            df = pd.DataFrame(
                {
                    "DATE": [datetime.now()],
                    "JOB": [self.job_name],
                    "STATUS": ["fail"],
                    "CHECKPOINT": [checkpoint_name],
                    "MESSAGE": [message],
                    "DATA": [f"{data}"],
                    "EXECUTION_MODE": [execution_mode],
                }
            )
            report_df = session.createDataFrame(df)
            report_df.write.mode("append").save_as_table("SNOWPARK_CHECKPOINTS_REPORT")

    def mark_pass(self, checkpoint_name, execution_mode=SCHEMA_EXECUTION_MODE):
        if self.log_results:
            session = self.snowpark_session
            df = pd.DataFrame(
                {
                    "DATE": [datetime.now()],
                    "JOB": [self.job_name],
                    "STATUS": ["pass"],
                    "CHECKPOINT": [checkpoint_name],
                    "MESSAGE": [""],
                    "DATA": [""],
                    "EXECUTION_MODE": [execution_mode],
                }
            )
            report_df = session.createDataFrame(df)
            report_df.write.mode("append").save_as_table("SNOWPARK_CHECKPOINTS_REPORT")



#
# Copyright (c) 2012-2024 Snowflake Computing Inc. All rights reserved.
#

from typing import Optional

from snowflake.snowpark_checkpoints.job_context import SnowparkJobContext


class SnowparkCheckpointError(Exception):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name: str,
        data=None,
    ):
        job_name = job_context.job_name if job_context else "Unknown Job"
        super().__init__(
            f"Job: {job_name} Checkpoint: {checkpoint_name}\n{message} \n {data}"
        )
        if job_context:
            job_context.mark_fail(
                message,
                checkpoint_name,
                data,
            )


class SparkMigrationError(SnowparkCheckpointError):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name=None,
        data=None,
    ):
        super().__init__(message, job_context, checkpoint_name, data)


class SchemaValidationError(SnowparkCheckpointError):
    def __init__(
        self,
        message,
        job_context: Optional[SnowparkJobContext],
        checkpoint_name=None,
        data=None,
    ):
        super().__init__(message, job_context, checkpoint_name, data)
